/*
 * modbus-utils - ping.c
 * 	version 0.1.0
 * 
 * Copyright (C) 2010  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <unistd.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <sys/signal.h>
#include <stdarg.h>
#include <syslog.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/poll.h>
#include <limits.h>
#include <wrapper.h>

#include <list.h>
#include <modbus.h>
#include <mclock.h>
#include <compiler.h>

int silent_level = 1;
static const char *modbus_device = "/dev/modbus";
static struct modbus_addr addr = MODBUS_ADDR_NULL;
static volatile int term;
	
static unsigned long long min_t = 0;
static unsigned long long max_t = 0;
static unsigned long long sum_t = 0;
static unsigned long count = 0;

static void show_usage(char *name)
{
	printf("modbus-ping " PACKAGE_VERSION ".\n"
"Copyright (C) 2010  Krzysztof Mazur <krzysiek@podlesie.net>.\n");

	printf("usage: %s [options]\n"
			"Options:\n"
"      --syslog			use syslog\n"
"  -V, --version\t		show version\n"
"  -d, --device=DEV		use selected ModBUS device\n"
"  -h, --help			show this help\n"
"  -q, --quiet			be quiet\n"
"  -v, --verbose\t		be verbose\n"
			"\n"
			"Report bugs to krzysiek@podlesie.net.\n",
			(name == NULL) ? "modbus-ping" : name);
	exit(EXIT_SUCCESS);
}

static void show_version(void)
{
	fprintf(stderr, "modbus-ping " PACKAGE_VERSION ".\n"
"Copyright (C) 2010  Krzysztof Mazur <krzysiek@podlesie.net>.\n");

	fprintf(stderr,	"This program comes with ABSOLUTELY NO WARRANTY; "
			"see COPYING for details.\n"
			"This is free software, and you are welcome to "
			"redistribute it\n"
			"under certain conditions; see COPYING for details.\n");
	exit(EXIT_SUCCESS);
}

static void parse_command_line(int argc, char **args)
{
	int option_index = 0;
	int ret;
	int c;

	static struct option long_options[] = {
		{"device", 1, 0, 'd'},
		{"help", 0, 0, 'h'},
		{"quiet", 0, 0, 'q'},
		{"syslog", 0, 0, 1},
		{"verbose", 0, 0, 'v'},
		{"version", 0, 0, 'V'},
		{0, 0, 0, 0},
	};

	while (1) {
		c = getopt_long(argc, args, "d:hl:n:r:qvV",
				long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case 'd':
			modbus_device = optarg;
			break;
		case 'h':
			show_usage(args[0]);
			BUG();
		case 'v':
			silent_level++;
			break;
		case 'q':
			silent_level--;
			break;
		case 'V':
			show_version();
			BUG();
		default:
			fprintf(stderr, "Unknown option.\n");
			exit(EXIT_FAILURE);
		}
	}
	
	if (optind < argc) {
		ret = modbus_name_to_addr(&addr, args[optind]);
		if (ret) {
			fprintf(stderr, "invalid address: %s", args[optind]);
			exit(EXIT_FAILURE);
		}
		optind++;

		if (optind < argc)
			fprintf(stderr, "warning: unknown parameters "
					"detected.\n");
	}
}

void SIGINT_handler(int arg, siginfo_t *siginfo, void *p)
{
	term = 1;
}

void SIGTERM_handler(int arg, siginfo_t *siginfo, void *p)
{
	term = 1;
}

static volatile int stats;

void SIGQUIT_handler(int arg, siginfo_t *siginfo, void *p)
{
	stats = 1;
}

void show_stats(void)
{
	fprintf(stderr, "min/avg/max = %.3f/%.3f/%.3f\n",
			min_t / 1000000.0,
			sum_t / (1000000.0 * count),
			max_t / 1000000.0);
	stats = 0;
}

int main(int argc, char **args)
{
	int ret;
	struct modbus_context ctx;
	unsigned long long t;
	char buf[2];
	struct sigaction action;

	parse_command_line(argc, args);
	ret = modbus_init(&ctx);
	if (ret)
		die("modbus_init: %s\n", strerror(-ret));
	ret = modbus_open(&ctx, modbus_device);
	if (ret)
		die("modbus_init: %s\n", strerror(-ret));

	mclock_init();
	
	action.sa_sigaction = SIGTERM_handler;
	sigemptyset(&action.sa_mask);
	action.sa_flags = SA_SIGINFO;
	sigaction(SIGTERM, &action, NULL);
	
	action.sa_sigaction = SIGINT_handler;
	sigemptyset(&action.sa_mask);
	action.sa_flags = SA_SIGINFO;
	sigaction(SIGINT, &action, NULL);
	
	action.sa_sigaction = SIGQUIT_handler;
	sigemptyset(&action.sa_mask);
	action.sa_flags = SA_SIGINFO;
	sigaction(SIGQUIT, &action, NULL);

	while (!term) {
		t = time_ns();
		ret = modbus_read(&ctx, &addr, &buf, 2, 0);
		t = time_ns() - t;
		if (stats)
			show_stats();
		if (ret == -EFAULT || ret > 0) {
			fprintf(stderr, "%s: time=%.2f ms\n",
					"dev", t / 1000000.0);
			if (!count)
				min_t = t;
			count++;
			if (t < min_t)
				min_t = t;
			if (t > max_t)
				max_t = t;
			sum_t += t;
		}
		sleep(1);
		if (stats)
			show_stats();
	}

	show_stats();

	modbus_close(&ctx);
	modbus_destroy(&ctx);
	return 0;
}
