/*
 * modbus-utils - tcp-master.c
 *
 * Modbus/TCP Master
 *
 * Copyright (C) 2013  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <unistd.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <sys/signal.h>
#include <stdarg.h>
#include <syslog.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/poll.h>
#include <modbus.h>
#include <compiler.h>
#include <rejd_byteorder.h>
#include <wrapper.h>
#include <mclock.h>
#include <pu-buff.h>

#define MODBUS_SLAVE_MAX	255

#define MODBUS_READ_HOLDING	0x03
#define MODBUS_READ_INPUT	0x04
#define MODBUS_WRITE_HOLDING	0x10

#define IO_BLOCKS	32

#include <list.h>

#if CONFIG_DEBUG
#define debug(fmt...) fprintf(stderr, fmt)
#else
#define debug(fmt...) do { } while (0)
#endif

#define ERROR_LIMIT	10

int verbose_level = 1;
static int use_syslog;

static const char *server_host = "localhost";
static const char *server_port = "502";
static const char *config_file;

struct ioblock {
	int op;
	int fd;
	unsigned int offset;

	unsigned int seq;
	unsigned int slave;
	unsigned int addr;
	unsigned int len;
	unsigned int period;

	unsigned int errors;

	unsigned long long next_update;

	struct pu_buff *send;
	struct pu_buff *recv;
};
static struct ioblock ioblock[IO_BLOCKS];
static unsigned int ioblock_n;
static unsigned int ioblock_current;

void mlog(int priority, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	if (use_syslog)
		vsyslog(priority, fmt, ap);
	else
		vfprintf(stderr, fmt, ap);
	va_end(ap);
}

static void show_usage(char *name)
{
	printf("modbus-tcp-master " PACKAGE_VERSION ".\n"
"Copyright (C) 2013  Krzysztof Mazur <krzysiek@podlesie.net>.\n");

	printf("usage: %s [options]\n"
			"Options:\n"
"      --syslog			use syslog\n"
"  -c, --config=FILE		configuraton file\n"
"  -V, --version\t		show version\n"
"  -h, --help			show this help\n"
"  -p, --port=PORT		use specified TCP port (502)\n"
"  -q, --quiet			be quiet\n"
"  -v, --verbose\t		be verbose\n"
			"\n"
			"Report bugs to krzysiek@podlesie.net.\n",
			(name == NULL) ? "modbus-tcp-master" : name);
	exit(EXIT_SUCCESS);
}

static void show_version(void)
{
	fprintf(stderr, "modbus-tcp-master " PACKAGE_VERSION ".\n"
"Copyright (C) 2013  Krzysztof Mazur <krzysiek@podlesie.net>.\n");

	fprintf(stderr,	"This program comes with ABSOLUTELY NO WARRANTY; "
			"see COPYING for details.\n"
			"This is free software, and you are welcome to "
			"redistribute it\n"
			"under certain conditions; see COPYING for details.\n");
	exit(EXIT_SUCCESS);
}

static void parse_command_line(int argc, char **args)
{
	static const struct option long_options[] = {
		{"config", 1, 0, 'c'},
		{"help", 0, 0, 'h'},
		{"port", 1, 0, 'p'},
		{"quiet", 0, 0, 'q'},
		{"syslog", 0, 0, 1},
		{"verbose", 0, 0, 'v'},
		{"version", 0, 0, 'V'},
		{0, 0, 0, 0},
	};
	int option_index = 0;
	int c;

	while (1) {
		c = getopt_long(argc, args, "c:hp:qvV",
				long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case 1:
			use_syslog = 1;
			break;
		case 'c':
			config_file = strdup(optarg);
			break;
		case 'h':
			show_usage(args[0]);
			BUG();
		case 'v':
			verbose_level++;
			break;
		case 'p':
			server_port = strdup(optarg);
			break;
		case 'q':
			verbose_level--;
			break;
		case 'V':
			show_version();
			BUG();
		default:
			fprintf(stderr, "Unknown option.\n");
			exit(EXIT_FAILURE);
		}
	}

	if (optind >= argc)
		return;

	server_host = strdup(args[optind++]);
	if (optind >= argc)
		return;

	fprintf(stderr, "warning: unknown parameters detected.\n");
}

static int do_connect(const char *host, const char *port)
{
	struct addrinfo hints, *ai, *aitop;
	int ret;
	int fd;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = 0;
	ret = getaddrinfo(host, port, &hints, &aitop);
	if (ret < 0) {
		fprintf(stderr, "getaddrinfo(%s, %s): %s\n", host, port,
				strerror(errno));
		exit(EXIT_FAILURE);
	}

	for (ai = aitop; ai; ai = ai->ai_next) {
		fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
		if (fd == -1) {
			fprintf(stderr, "warning: socket: %s\n",
					strerror(errno));
			continue;
		}

		ret = connect(fd, ai->ai_addr, ai->ai_addrlen);
		if (ret < 0) {
			fprintf(stderr, "warning: connect: %s\n",
					strerror(errno));
			close(fd);
			continue;
		}
		freeaddrinfo(aitop);
		return fd;
	}
	fprintf(stderr, "error: can't connect to %s:%s\n", host, port);
	exit(EXIT_FAILURE);
}

struct modbus_03_cmd {
	uint8_t cmd;			/* command */
	uint16_t addr;			/* address high byte */
	uint16_t len;			/* length high byte, words */
} __packed;

struct modbus_03_resp {
	uint8_t cmd;			/* command */
	uint8_t bytes;			/* big endian, payload bytes */
	uint16_t payload[0];
} __packed;

/* 0x10 command - Write multiple registers */
struct modbus_10_cmd {
	uint8_t cmd;			/* command */
	uint16_t addr;			/* address */
	uint16_t len;			/* length, words */
	uint8_t bytes;			/* payload length */
	uint16_t payload[0];
} __packed;

struct modbus_10_resp {
	uint8_t cmd;			/* command */
	uint8_t addr_hi;		/* address high byte */
	uint8_t addr_lo;		/* address low byte */
	uint8_t len_hi;			/* length high byte, words */
	uint8_t len_lo;			/* length low byte, words */
} __packed;

struct modbus_err_msg {
	uint8_t cmd;			/* command */
	uint8_t err;			/* exception */
} __packed;

static struct pu_buff *modbus_alloc(unsigned int size, pu_gfp_t priority)
{
	struct pu_buff *p;

	p = pu_alloc_skb(size + sizeof(struct mbap_header), priority);
	if (p)
		pu_skb_reserve(p, sizeof(struct mbap_header));
	return p;
}

static struct pu_buff *ioblock_prepare_read(struct ioblock *iob, int op)
{
	struct modbus_03_cmd *cmd;
	struct pu_buff *skb;

	skb = modbus_alloc(sizeof(*cmd), GFP_KERNEL);
	if (!skb)
		return skb;

	cmd = (void *) pu_skb_put(skb, sizeof(*cmd));
	cmd->cmd = op;
	cmd->addr = cpu_to_be16(iob->addr);
	cmd->len = cpu_to_be16(iob->len);
	return skb;
}

static struct pu_buff *ioblock_prepare_write(struct ioblock *iob, int op)
{
	struct modbus_10_cmd *cmd;
	struct pu_buff *skb;

	skb = modbus_alloc(sizeof(*cmd) + 2 * iob->len, GFP_KERNEL);
	if (!skb)
		return skb;

	cmd = (void *) pu_skb_put(skb, sizeof(*cmd) + 2 * iob->len);
	cmd->cmd = op;
	cmd->addr = cpu_to_be16(iob->addr);
	cmd->len = cpu_to_be16(iob->len);
	cmd->bytes = 2 * iob->len;
	pread(iob->fd, cmd + 1, 2 * iob->len, iob->offset);
	return skb;
}

static int ioblock_gen_request(struct ioblock *iob)
{
	struct pu_buff *skb;
	struct mbap_header *hdr;

	switch (iob->op) {
	case MODBUS_READ_HOLDING:
	case MODBUS_READ_INPUT:
		skb = ioblock_prepare_read(iob, iob->op);
		break;
	case MODBUS_WRITE_HOLDING:
		skb = ioblock_prepare_write(iob, iob->op);
		break;
	default:
		return -EINVAL;
	}
	hdr = (void *) pu_skb_push(skb, sizeof(*hdr));
	if (!hdr)
		return -ENOMEM;

	hdr->seq = cpu_to_be16(iob - ioblock);
	hdr->proto = 0;
	hdr->len = cpu_to_be16(skb->len + 1 - sizeof(*hdr));
	hdr->addr = 0;
	iob->send = skb;
	return 0;
}

static int ioblock_try_xmit(int fd, struct ioblock *iob)
{
	struct pu_buff *skb = iob->send;
	int ret = 0;

	if (!skb)
		return 0;

	if (skb->len) {
		ret = xwrite(fd, skb->data, skb->len);
		if (ret < 0)
			return -errno;
		if (!ret)
			return -ECONNRESET;
		pu_skb_pull(skb, ret);
	}
	return 0;
}

static struct pu_buff *pu_buff_read(int fd, struct pu_buff *skb, size_t count)
{
	unsigned char *p;
	ssize_t ret;

	if (pu_skb_tailroom(skb) < count) {
		struct pu_buff *skb2 = skb_copy_expand(skb, 0, count,
				GFP_KERNEL);
		pu_consume_skb(skb);
		if (skb == NULL)
			return skb;
		skb = skb2;
	}
	p = pu_skb_put(skb, 0);

	ret = xread(fd, p, count);
	if (ret > 0)
		pu_skb_put(skb, ret);
	else if (ret != -EAGAIN)
		die("read: %s", strerror(-ret));
	return skb;
}

static void modbus_recv(struct pu_buff *skb)
{
	struct mbap_header *hdr = (void *) skb->data;
	unsigned int seq = be16_to_cpu(hdr->seq);
	struct ioblock *iob;
	uint8_t *payload;
	int plen;
	int ret;

	pu_skb_pull(skb, sizeof(*hdr));
	if (seq > ARRAY_SIZE(ioblock))
		die("invalid sequence number %d", seq);
	iob = &ioblock[seq];

	if (!iob->send)
		die("unexpected sequence number: %d", seq);

	payload = (void *) skb->data;
	if (payload[0] != iob->op) {
		die("invalid command, 0x%02x should be 0x%02x",
				payload[0], iob->op);
	}

	switch (iob->op) {
	case MODBUS_READ_HOLDING:
	case MODBUS_READ_INPUT:
		plen = 2 * iob->len;
		ret = pwrite(iob->fd, payload + 2, plen, iob->offset);
		if ((ret < plen) && (iob->errors < ERROR_LIMIT)) {
			warn("%d: pwrite error: %d: %s", seq, ret,
					(ret < 0) ? strerror(errno)
					: "short write");
			iob->errors++;
		}
		break;
	case MODBUS_WRITE_HOLDING:
		break;
	default:
		warn("Unknown operation: %d", iob->op);
	}

	pu_consume_skb(skb);

	pu_consume_skb(iob->send);
	iob->send = NULL;
}

static struct pu_buff *recv_skb;

static int frame_read(int fd)
{
	struct mbap_header *hdr;
	unsigned long len;
	int nowait = 0;

	if (!recv_skb)
		recv_skb = pu_alloc_skb(MBAP_MTU, GFP_KERNEL);

	if (recv_skb->len < sizeof(*hdr)) {
		recv_skb = pu_buff_read(fd, recv_skb, sizeof(*hdr));
		if (!recv_skb)
			goto err_close_conn;
		if (recv_skb->len < sizeof(*hdr))
			return 0;
		nowait = 1;
	}

	hdr = (void *) recv_skb->data;
	if (hdr->proto != 0) {
		die("protocol error: protocol = %04x\n", hdr->proto);
		goto err_close_conn;
	}

	/*
	 * we don't care about packets with error code
	 */
	len = be16_to_cpu(hdr->len);
	if (len >= MODBUS_ADU_SIZE_MAX + 1) {
		die("protocol error: too large message %ld\n", len);
		goto err_close_conn;
	}
	len += sizeof(*hdr) - 1;

	if (len - recv_skb->len > 0) {
		if (nowait)
			return 0;

		recv_skb = pu_buff_read(fd, recv_skb, len - recv_skb->len);
		if (!recv_skb)
			goto err_close_conn;
	}

	if (recv_skb->len < len)
		return 0;

	modbus_recv(recv_skb);
	recv_skb = NULL;
	return 0;

err_close_conn:
	return -ECONNRESET;
}


static int main_loop(int fd)
{
	struct pollfd pfd[1];
	unsigned long long t;
	long timeout;
	unsigned int i;
	int ret;

	pfd[0].fd = fd;
	for (;;) {
		t = time_ms();
		timeout = 1000;
		pfd[0].events = POLLIN;
		for (i = 0; i < ioblock_n; i++) {
			struct ioblock *iob = &ioblock[i];
			long diff;

			if (!iob->send) {
				diff = iob->next_update - t;

				if (diff <= 0) {
					ret = ioblock_gen_request(iob);
					if (ret) {
						die("ioblock_gen: %s",
								strerror(-ret));
					}
					iob->next_update += iob->period;
				}
				if (diff <= timeout)
					timeout = diff;
				/* FIXME: add abort */
			}

			if (iob->send && iob->send->len)
				pfd[0].events |= POLLOUT;
		}

		if (timeout < 0)
			timeout = 0;

		ret = poll(pfd, 1, timeout);
		if (ret < 0) {
			if (errno == EINTR)
				continue;
			die_errno("poll error");
		}
		if (!ret)
			continue;

		if (pfd[0].revents & (POLLERR | POLLHUP | POLLNVAL)) {
			die("poll: file descriptor error: %08x",
					pfd[0].revents);
		}

		if (pfd[0].revents & POLLIN) {
			ret = frame_read(fd);
			if (ret < 0)
				die("frame_read error");
		}

		if (pfd[0].revents & POLLOUT) {
			for (i = 0; i < ioblock_n; i++) {
				struct ioblock *iob = &ioblock[ioblock_current];

				ret = ioblock_try_xmit(fd, iob);
				if (ret < 0) {
					if (ret == -EAGAIN)
						break;
				}

				ioblock_current = ioblock_current + 1;
				if (ioblock_current >= ioblock_n)
					ioblock_current = 0;
			}
		}
	}
	return 0;
}

static int parse_op(const char *op)
{
	if (op == NULL)
		return -EINVAL;

	if (!strcmp(op, "read"))
		return MODBUS_READ_HOLDING;
	else if (!strcmp(op, "read-input"))
		return MODBUS_READ_INPUT;
	else if (!strcmp(op, "write"))
		return MODBUS_WRITE_HOLDING;
	return -EINVAL;
}

static int add_ioblock(const char *op, const char *dev,
		unsigned int offset, unsigned int slave,
		unsigned int addr, unsigned int length,
		unsigned int period)
{
	struct ioblock *b = &ioblock[ioblock_n];
	int flags;
	int ret;
	int fd;

	ret = -ENFILE;
	if (ioblock_n >= ARRAY_SIZE(ioblock)) {
		warn("maximal I/O blocks count limit reached");
		goto out;
	}

	memset(b, 0, sizeof(b));

	ret = -EINVAL;
	b->op = parse_op(op);
	if (b->op < 0) {
		warn("invalid operation '%s'", op ? op : "");
		goto out;
	}

	if (!dev)
		goto out;
	flags = (b->op == MODBUS_WRITE_HOLDING) ? O_RDONLY : O_WRONLY;
	fd = open(dev, flags);
	if (fd == -1) {
		warn("cannot open %s: %s", dev, strerror(errno));
		goto out;
	}
	b->fd = fd;

	if (offset < 0) {
		warn("invalid offset %d", offset);
		goto out;
	}
	b->offset = offset;

	if (slave < 0 || slave > MODBUS_SLAVE_MAX) {
		warn("invalid slave address %d", slave);
		goto out;
	}
	b->slave = slave;

	if (addr < 0 || addr > 65535) {
		warn("invalid address %d", addr);
		goto out;
	}
	b->addr = addr;

	if (length < 0 || length > 100) {
		warn("invalid length %d", length);
		goto out;
	}
	b->len = length;

	if (period <= 0)
		period = 100;
	b->period = period;

	b->next_update = time_ms();
	ioblock_n++;
	ret = 0;

out:
	xfree(op);
	xfree(dev);
	return ret;
}

static void parse_config(FILE *f)
{
	char *line = NULL;
	size_t n = 0;
	ssize_t len;
	int ret;

	/*
	 * configuration file:
	 * OP	DEVICE	OFFSET	SLAVE	ADDR	LEN	PERIOD
	 */
	while ((len = getline(&line, &n, f)) != -1) {
		char *op = NULL;
		char *dev = NULL;
		unsigned int offset = 0;
		unsigned int slave = 0;
		unsigned int addr = 0;
		unsigned int length = 0;
		unsigned int period = 100;

		if (line[0] == '#')
			continue;

		ret = sscanf(line, "%ms %ms %d %d %d %d %d", &op, &dev,
				&offset, &slave, &addr, &length, &period);
		if (ret < 1) {
			warn("empty line?");
			continue;
		}

		ret = add_ioblock(op, dev, offset, slave, addr, length,
				period);
		if (ret < 0)
			warn("add_ioblock: %s", strerror(-ret));
	}
	xfree(line);
}

int main(int argc, char **args)
{
	FILE *f = stdin;
	int ret;
	int fd;
	int op;

	signal(SIGPIPE, SIG_IGN);
	parse_command_line(argc, args);

	mclock_init();
	if (config_file) {
		f = fopen(config_file, "r");
		if (f == NULL)
			die_errno("fopen(%s)", config_file);
	}
	parse_config(f);
	if (config_file)
		fclose(f);

	if (use_syslog)
		openlog("modbus-tcp-master", LOG_PID, LOG_DAEMON);

	fd = do_connect(server_host, server_port);

	op = 1;
	ret = setsockopt(fd, SOL_TCP, TCP_NODELAY, &op, sizeof(op));
	if (ret)
		warn("setsockopt(TCP_NODELAY): %s", strerror(errno));

	ret = setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, &op, sizeof(op));
	if (ret)
		warn("setsockopt(TCP_KEEPALIVE): %s", strerror(errno));

	ret = fcntl(fd, F_SETFL, O_NONBLOCK);
	if (ret)
		warn("fcntl(): %s", strerror(errno));

	main_loop(fd);
	return 0;
}
