/*
 * modbus-tcp - main.c
 *	version 0.1.0
 *
 *	ModBUS TCP slave
 *
 * Copyright (C) 2005..2010  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <unistd.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <sys/signal.h>
#include <stdarg.h>
#include <syslog.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/poll.h>
#include <modbus.h>
#include <compiler.h>
#include <rejd_byteorder.h>
#include <wrapper.h>

#include <list.h>
#include <context.h>

#if CONFIG_DEBUG
#define debug(fmt...) fprintf(stderr, fmt)
#else
#define debug(fmt...) do { } while (0)
#endif

int verbose_level = 1;
static int use_syslog = 0;
static const char *modbus_device = "/dev/modbus";
static unsigned int rlimit = 1;
static int bus_id = 0;
static int emu = 0;

struct modbus_context modbus;

static int lport = 502;	/* listen on port */
static int max_conns = 20;

void mlog(int priority, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	if (use_syslog)
		vsyslog(priority, fmt, ap);
	else
		vfprintf(stderr, fmt, ap);
	va_end(ap);
}

static void show_usage(char *name)
{
	printf("modbus-tcp " PACKAGE_VERSION ".\n"
"Copyright (C) 2005..2010  Krzysztof Mazur <krzysiek@podlesie.net>.\n");

	printf("usage: %s [options]\n"
			"Options:\n"
"      --syslog			use syslog\n"
"  -b, --bus-id\t		modbus bus number\n"
"  -V, --version\t		show version\n"
"  -d, --device=DEV		use selected Modbus device\n"
"  -e, --emu			use builtin Modbus implementation\n"
"  -h, --help			show this help\n"
"  -l, --lport=PORT		listen port (502)\n"
"  -n, --conn-limit=LIMIT	connections limit(%d)\n"
"  -q, --quiet			be quiet\n"
"  -r, --rlimit=RLIMIT		request limit per client (8)\n"
"  -v, --verbose\t		be verbose\n"
			"\n"
			"Report bugs to krzysiek@podlesie.net.\n",
			(name == NULL) ? "modbus-tcp" : name, max_conns);
	exit(EXIT_SUCCESS);
}

static void show_version(void)
{
	fprintf(stderr, "modbus-tcp " PACKAGE_VERSION ".\n"
"Copyright (C) 2005..2010  Krzysztof Mazur <krzysiek@podlesie.net>.\n");

	fprintf(stderr,	"This program comes with ABSOLUTELY NO WARRANTY; "
			"see COPYING for details.\n"
			"This is free software, and you are welcome to "
			"redistribute it\n"
			"under certain conditions; see COPYING for details.\n");
	exit(EXIT_SUCCESS);
}

static void parse_command_line(int argc, char **args)
{
	int c;
	int option_index = 0;

	static struct option long_options[] = {
		{"conn-limit", 1, 0, 'n'},
		{"bus-id", 1, 0, 'b'},
		{"device", 1, 0, 'd'},
		{"emu", 0, 0, 'e'},
		{"help", 0, 0, 'h'},
		{"lport", 1, 0, 'l'},
		{"quiet", 0, 0, 'q'},
		{"rlimit", 1, 0, 'r'},
		{"syslog", 0, 0, 1},
		{"verbose", 0, 0, 'v'},
		{"version", 0, 0, 'V'},
		{0, 0, 0, 0},
	};

	while (1) {
		c = getopt_long(argc, args, "b:d:ehl:n:r:qvV",
				long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case 1:
			use_syslog = 1;
			break;
		case 'b':
			bus_id = atoi(optarg);
			break;
		case 'd':
			modbus_device = optarg;
			break;
		case 'e':
			emu = 1;
			break;
		case 'h':
			show_usage(args[0]);
			BUG();
		case 'l':
			lport = atoi(optarg);
			break;
		case 'n':
			max_conns = atoi(optarg);
			break;
		case 'v':
			verbose_level++;
			break;
		case 'r':
			rlimit = atoi(optarg);
			break;
		case 'q':
			verbose_level--;
			break;
		case 'V':
			show_version();
			BUG();
		default:
			fprintf(stderr, "Unknown option.\n");
			exit(EXIT_FAILURE);
		}
	}

	if (optind < argc)
		fprintf(stderr, "warning: unknown parameters detected.\n");
}

static int listen_fd = -1;

static int do_listen(void)
{
	struct sockaddr_in serv_addr;
	int fd;
	int op;

	if (verbose_level > 1)
		fprintf(stderr, "Listening on port %d.\n", lport);
	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
		die("socket(): %s.\n", strerror(errno));
	memset(&serv_addr, 0, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	serv_addr.sin_port = htons(lport);
	op = 1;
	setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &op, sizeof(op));
	if (bind(fd, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) == -1)
		die("bind(): %s.\n", strerror(errno));
	if (listen(fd, 5))
		die("listen(): %s.\n", strerror(errno));
	listen_fd = fd;
	return 0;
}

static LIST_HEAD(contexts);
int conn_count;

static void accept_connection(void)
{
	int ret;
	struct c_context *c;
	struct sockaddr_in addr;
	socklen_t addrlen = sizeof(addr);
	int op;

	ret = accept(listen_fd, &addr, &addrlen);
	if (ret < 0) {
		fprintf(stderr, "accept error: %s", strerror(errno));
		return;
	}
	if (conn_count >= max_conns) {
		fprintf(stderr, "accept: connections limit exceed.\n");
		close(ret);
		return;
	}

	if (verbose_level > 0)
		fprintf(stderr, "accepted connection(%d): fd = %d\n",
				conn_count, ret);

	if (fcntl(ret, F_SETFL, O_NONBLOCK)) {
		fprintf(stderr, "fcntl(): %s", strerror(errno));
		close(ret);
		return;
	}

	op = 1;
	setsockopt(ret, SOL_TCP, TCP_NODELAY, &op, sizeof(op));
	setsockopt(ret, SOL_SOCKET, SO_KEEPALIVE, &op, sizeof(op));

	c = xmalloc(sizeof(*c));
	context_init(c, ret);
	list_add_tail(&c->list, &contexts);
	conn_count++;
}

static int main_loop(void)
{
	struct pollfd *pfd;
	int ret;
	int cfd;
	struct c_context *c;
	struct list_head *t;

	pfd = xmalloc((max_conns + 3) * sizeof(struct pollfd));
	for (;;) {
		orphan_killer();

		/* prepare for poll */
		pfd[0].fd = listen_fd;
		pfd[0].events = POLLIN;
		cfd = 1;

		if (!emu) {
			/* modbus */
			pfd[1].fd = modbus.fd;
			pfd[1].events = POLLIN;
			cfd++;
		}

		list_for_each_entry(c, &contexts, list) {
			pfd[cfd].fd = c->fd;
			pfd[cfd].events = 0;
			if (c->rcount < rlimit)
				pfd[cfd].events |= POLLIN;
			if (!list_empty(&c->req_ready))
				pfd[cfd].events |= POLLOUT;
			cfd++;
		}

		ret = poll(pfd, cfd, -1);
		if (ret < 0) {
			if (errno == EINTR)
				continue;
			die("poll() error: %s\n", strerror(errno));
		} else if (!ret)
			continue;

		cfd = emu ? 1 : 2;
		for (t = list_next(&contexts); t != &contexts;) {
			c = list_entry(t, struct c_context, list);
			t = list_next(t);

			BUG_ON(c->fd != pfd[cfd].fd);
			if (pfd[cfd].revents & (POLLERR | POLLHUP | POLLNVAL)) {
				context_destroy(c);
			} else if (c->write_wait && !c->rcount) {
				context_destroy(c);
			} else {
				ret = 0;
				if (pfd[cfd].revents & pfd[cfd].events & POLLIN)
					ret = context_read(c);

				if ((pfd[cfd].revents & POLLOUT)
						&& (ret != -ECONNRESET)) {
					ret = context_write(c);
					if (ret == -ECONNRESET)
						context_destroy(c);
				}
			}
			cfd++;
		}

		if (pfd[0].revents & POLLIN)
			accept_connection();

		if (emu || (pfd[1].revents & POLLIN)) {
			struct modbus_request *mr;
			struct c_request *r;

			do {
				ret = modbus_process(&modbus, &mr, 0);
				if (!ret) {
					struct mbap_header *hdr;

					r = container_of(mr, struct c_request,
							mreq);
					list_del(&r->list);
					if (r->ctx != NULL) {
						list_add(&r->list,
							&r->ctx->req_ready);
						hdr = mr->r;
						hdr->seq = cpu_to_be16(
								r->tcp_id);
						hdr->addr = r->tcp_addr;
					} else {
						request_destroy(r);
					}
				} else {
					break;
				}
			} while (emu);
		}
	}
	return 0;
}

void initialize_modbus(void)
{
	int ret;

	ret = modbus_init(&modbus);
	if (ret) {
		fprintf(stderr, "modbus initialization error\n");
		exit(EXIT_FAILURE);
	}
	ret = modbus_set_emu(&modbus, emu);
	if (ret) {
		fprintf(stderr, "modbus: set_emu: %s\n", strerror(-ret));
		exit(EXIT_FAILURE);
	}
	ret = modbus_open(&modbus, modbus_device);
	if (ret) {
		fprintf(stderr, "modbus: open: %s\n", strerror(-ret));
		exit(EXIT_FAILURE);
	}
	modbus_set_packet(&modbus);
	modbus_set_bus(&modbus, bus_id);
}

int main(int argc, char **args)
{
	signal(SIGPIPE, SIG_IGN);
	parse_command_line(argc, args);

	initialize_modbus();
	if (use_syslog)
		openlog("modbus-tcp", LOG_PID, LOG_DAEMON);
	do_listen();
	main_loop();
	return 0;
}
